# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#      :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
from __future__ import annotations

from copy import copy
from math import sqrt
from unittest import TestCase

import pytest
from gemseo import execute_algo
from gemseo.algos.design_space import DesignSpace
from gemseo.algos.opt.factory import OptimizationLibraryFactory
from gemseo.algos.optimization_problem import OptimizationProblem
from gemseo.core.mdo_functions.mdo_function import MDOFunction
from gemseo.problems.optimization.rosenbrock import Rosenbrock
from gemseo.utils.testing.opt_lib_test_base import OptLibraryTestBase
from numpy import isnan
from numpy import nan
from scipy.optimize import rosen
from scipy.optimize import rosen_der

from gemseo_pdfo.pdfo import PDFOOpt
from gemseo_pdfo.settings.pdfo_cobyla_settings import PDFO_COBYLA_Settings

pytest.importorskip("pdfo", reason="pdfo is not available")


class TestPDFO(TestCase):
    OPT_LIB_NAME = "PDFOOpt"

    def test_failed(self):
        """"""
        algo_name = "PDFO_COBYLA"
        self.assertRaises(
            ValueError,
            OptLibraryTestBase.generate_error_test,
            "PDFOAlgorithms",
            algo_name=algo_name,
            max_iter=10,
            ensure_bounds=False,
        )

    def test_nan_handling(self):
        """Test that an occurrence of NaN value in the objective function does not stop
        the optimizer.

        In this case, a NaN "bubble" is put at the beginning of the optimizer path. In
        this test, it is expected that the optimizer will encounter and by-pass the NaN
        bubble.
        """
        opt_problem = Rosenbrock()

        fun = copy(opt_problem.objective)

        def wrapped_fun(x_vec):
            x = x_vec[0]
            y = x_vec[1]

            d = sqrt((x - 0.1) ** 2 + (y - 0.1) ** 2)

            if d < 0.05:
                return nan
            return fun.func(x_vec)

        opt_problem.objective._func = wrapped_fun
        opt_problem.stop_if_nan = False

        pdfo_cobyla_settings = PDFO_COBYLA_Settings(
            max_iter=10000, rhobeg=0.1, rhoend=1e-6
        )

        opt_result = execute_algo(
            opt_problem, algo_type="opt", settings_model=pdfo_cobyla_settings
        )

        obj_history = opt_problem.database.get_function_history("rosen")

        is_nan = any(isnan(obj_history))
        assert is_nan
        assert opt_result.f_opt < 1e-3

    def test_nan_handling_2(self):
        """Test that an occurrence of NaN value in the objective function does not stop
        the optimizer.

        In this test, all the values of x>0.7 are not realizable. The optimum is then
        expected for x[0] ~= 0.7
        """
        opt_problem = Rosenbrock()

        fun = copy(opt_problem.objective)

        def wrapped_fun(x_vec):
            x = x_vec[0]

            if x > 0.7:
                return nan
            return fun.func(x_vec)

        opt_problem.objective._func = wrapped_fun
        opt_problem.stop_if_nan = False

        pdfo_cobyla_settings = PDFO_COBYLA_Settings(
            max_iter=10000, rhobeg=0.1, rhoend=1e-6
        )

        opt_result = execute_algo(
            opt_problem, algo_type="opt", settings_model=pdfo_cobyla_settings
        )

        obj_history = opt_problem.database.get_function_history("rosen")

        is_nan = any(isnan(obj_history))
        assert is_nan
        assert pytest.approx(opt_result.x_opt[0], rel=1e-3) == 0.7

    def test_xtol_ftol_activation(self):
        def run_pb(settings):
            design_space = DesignSpace()
            design_space.add_variable(
                "x1", 2, DesignSpace.DesignVariableType.FLOAT, -1.0, 1.0, 0.0
            )
            problem = OptimizationProblem(design_space)
            problem.objective = MDOFunction(rosen, "Rosenbrock", "obj", rosen_der)
            res = OptimizationLibraryFactory().execute(
                problem, algo_name="PDFO_COBYLA", **settings
            )
            return res, problem

        for tol_name in (
            "ftol_abs",
            "ftol_rel",
            "xtol_abs",
            "xtol_rel",
        ):
            res, pb = run_pb({tol_name: 1e10})
            assert tol_name in res.message
            # Check that the criteria is activated as ap
            assert len(pb.database) == 3


def get_options(algo_name):
    """
    :param algo_name:
    """
    return {"max_iter": 10000, "rhobeg": 0.3, "rhoend": 1e-6}


def get_pb_options(pb_name):
    """
    :param algo_name:
    """
    if pb_name == "Power2":
        return {"initial_value": 0.0}
    return {}


suite_tests = OptLibraryTestBase()
for test_method in suite_tests.generate_test("PDFOOpt", get_options, get_pb_options):
    setattr(TestPDFO, test_method.__name__, test_method)


def test_library_name():
    """Check the library name."""
    assert PDFOOpt.LIBRARY_NAME == "PDFO"



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""Settings for the optimization algorithms from PDFO."""

from __future__ import annotations

from functools import partial

from gemseo.algos.opt.base_optimizer_settings import BaseOptimizerSettings
from gemseo.utils.pydantic import copy_field
from numpy import inf
from pydantic import Field
from pydantic import NonNegativeFloat  # noqa:TC002
from pydantic import PositiveInt  # noqa:TC002

copy_field_opt = partial(copy_field, model=BaseOptimizerSettings)


class BasePDFOSettings(BaseOptimizerSettings):
    """The PDFO optimization library settings."""

    ftol_rel: NonNegativeFloat = copy_field_opt("ftol_rel", default=1e-12)

    ftol_abs: NonNegativeFloat = copy_field_opt("ftol_abs", default=1e-12)

    xtol_abs: NonNegativeFloat = copy_field_opt("xtol_abs", default=1e-12)

    xtol_rel: NonNegativeFloat = copy_field_opt("xtol_rel", default=1e-12)

    max_time: NonNegativeFloat = copy_field_opt("max_time", default=0)

    max_iter: PositiveInt = copy_field_opt("max_iter", default=500)

    rhobeg: NonNegativeFloat = Field(
        default=0.5, description="The initial value of the trust region radius."
    )

    rhoend: NonNegativeFloat = Field(
        default=1e-6,
        description=(
            "The final value of the trust region radius. "
            "Indicates the accuracy required in the final values of the variables."
        ),
    )

    ftarget: float = Field(
        default=-inf,
        description=(
            "The target value of the objective function. "
            "If a feasible iterate achieves an objective function value "
            "lower or equal to ``ftarget``, the algorithm stops immediately."
        ),
    )

    scale: bool = Field(
        default=False,
        description=(
            "The flag indicating whether to scale the problem "
            "according to the bound constraints."
        ),
    )

    quiet: bool = Field(
        default=True,
        description=(
            "The flag of quietness of the interface. "
            "If ``True``, the output message will not be printed."
        ),
    )

    classical: bool = Field(
        default=False,
        description=(
            "The flag indicating whether to call the classical Powell code or not."
        ),
    )

    debug: bool = Field(default=False, description="The debugging flag.")

    chkfunval: bool = Field(
        default=False,
        description=(
            "A flag used when debugging. "
            "If both ``debug`` and ``chkfunval`` are ``True``, "
            "an extra function/constraint evaluation is performed "
            "to check whether the returned values of the objective function "
            "and constraint match the returned x."
        ),
    )

    ensure_bounds: bool = Field(
        default=True,
        description=(
            "Whether to project the design vector onto the design space "
            "before execution."
        ),
    )

    normalize_design_space: bool = copy_field_opt(
        "normalize_design_space", default=True
    )



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""PDFO optimization library wrapper, see the [PDFO website](https://www.pdfo.net)."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Final

from gemseo.algos.design_space_utils import get_value_and_bounds
from gemseo.algos.opt.base_optimization_library import BaseOptimizationLibrary
from gemseo.algos.opt.base_optimization_library import OptimizationAlgorithmDescription
from gemseo.algos.opt.base_optimizer_settings import BaseOptimizerSettings
from numpy import isfinite
from numpy import ndarray
from numpy import real
from pdfo import pdfo

from gemseo_pdfo.settings.base_pdfo_settings import BasePDFOSettings
from gemseo_pdfo.settings.pdfo_bobyqa_settings import PDFO_BOBYQA_Settings
from gemseo_pdfo.settings.pdfo_cobyla_settings import PDFO_COBYLA_Settings
from gemseo_pdfo.settings.pdfo_newuoa_settings import PDFO_NEWUOA_Settings

if TYPE_CHECKING:
    from collections.abc import Callable

    from gemseo.algos.optimization_problem import OptimizationProblem

OptionType = str | int | float | bool | ndarray


@dataclass
class PDFOAlgorithmDescription(OptimizationAlgorithmDescription):
    """The description of an optimization algorithm from the PDFO library."""

    library_name: str = "PDFO"

    Settings: type[BasePDFOSettings] = BasePDFOSettings
    """The option valiation model for  PDFO optimization library."""


class PDFOOpt(BaseOptimizationLibrary):
    """PDFO optimization library interface.

    See OptimizationLibrary.
    """

    __DOC: Final[str] = "https://www.pdfo.net/"

    LIBRARY_NAME = PDFOAlgorithmDescription.library_name

    LIB_COMPUTE_GRAD = False

    ALGORITHM_INFOS: ClassVar[dict[str, Any]] = {
        "PDFO_COBYLA": PDFOAlgorithmDescription(
            algorithm_name="COBYLA",
            description="Constrained Optimization By Linear Approximations ",
            handle_equality_constraints=True,
            handle_inequality_constraints=True,
            internal_algorithm_name="cobyla",
            positive_constraints=True,
            website=f"{__DOC}",
            Settings=PDFO_COBYLA_Settings,
        ),
        "PDFO_BOBYQA": PDFOAlgorithmDescription(
            algorithm_name="BOBYQA",
            description="Bound Optimization By Quadratic Approximation",
            internal_algorithm_name="bobyqa",
            website=f"{__DOC}",
            Settings=PDFO_BOBYQA_Settings,
        ),
        "PDFO_NEWUOA": PDFOAlgorithmDescription(
            algorithm_name="NEWUOA",
            description="NEWUOA",
            internal_algorithm_name="newuoa",
            website=f"{__DOC}",
            Settings=PDFO_NEWUOA_Settings,
        ),
    }

    def __init__(self, algo_name: str) -> None:  # noqa: D107
        super().__init__(algo_name)
        self.name = "PDFO"

    def _run(self, problem: OptimizationProblem) -> tuple[str, Any]:
        # Get the normalized bounds:
        x_0, l_b, u_b = get_value_and_bounds(
            problem.design_space, self._settings.normalize_design_space
        )

        # Replace infinite values with None:
        l_b = [val if isfinite(val) else None for val in l_b]
        u_b = [val if isfinite(val) else None for val in u_b]

        ensure_bounds = self._settings.ensure_bounds

        cstr_pdfo = []
        for cstr in self._get_right_sign_constraints(problem):
            c_pdfo = {"type": cstr.f_type}
            if ensure_bounds:
                c_pdfo["fun"] = self.__ensure_bounds(cstr.func)
            else:
                c_pdfo["fun"] = cstr.func

            cstr_pdfo.append(c_pdfo)

        # Filter settings to get only the ones of the PDFO optimizer
        settings_ = self._filter_settings(
            self._settings.model_dump(), BaseOptimizerSettings
        )

        # |g| is in charge of ensuring max iterations, since it may
        # have a different definition of iterations, such as for SLSQP
        # for instance which counts duplicate calls to x as a new iteration
        settings_["maxfev"] = int(self._settings.max_iter * 1.2)

        def real_part_fun(
            x: ndarray,
        ) -> int | float:
            """Wrap the objective function and keep the real part.

            Args:
                x: The values to be given to the function.

            Returns:
                The real part of the evaluation of the function.
            """
            return real(problem.objective.evaluate(x))

        def ensure_bounds_fun(x_vect):
            return real_part_fun(
                self._problem.design_space.project_into_bounds(
                    x_vect, self._settings.normalize_design_space
                )
            )

        opt_result = pdfo(
            fun=ensure_bounds_fun if ensure_bounds else real_part_fun,
            x0=x_0,
            method=self.ALGORITHM_INFOS[self._algo_name].internal_algorithm_name,
            bounds=list(zip(l_b, u_b, strict=False)),
            constraints=cstr_pdfo,
            options=settings_,
        )

        return opt_result.message, opt_result.status

    def __ensure_bounds(
        self, orig_func: Callable[[ndarray], ndarray]
    ) -> Callable[[ndarray], ndarray]:
        """Project the design vector onto the design space before execution.

        Args:
            orig_func: The original function.

        Returns:
            A function calling the original function
            with the input data projected onto the design space.
        """

        def wrapped_func(x_vect):
            return orig_func(
                self._problem.design_space.project_into_bounds(
                    x_vect, self._settings.normalize_design_space
                )
            )

        return wrapped_func



# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import annotations

from pathlib import Path

file_dir_path = Path(__file__).parent
example_dir_name = "examples"
gallery_dir = file_dir_path / "generated" / example_dir_name
examples_dir = file_dir_path / example_dir_name
examples_subdirs = [
    subdir.name
    for subdir in examples_dir.iterdir()
    if (examples_dir / subdir).is_dir()
    and (examples_dir / subdir / "README.md").is_file()
]

conf = {
    f"{example_dir_name}_dirs": [examples_dir / subdir for subdir in examples_subdirs],
    "gallery_dirs": [gallery_dir / subdir for subdir in examples_subdirs],
}



# ISC License
#
# Copyright (c) 2021, Timothée Mazzucotelli
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Adapted from https://github.com/mkdocstrings/griffe/blob/main/scripts/gen_credits.py
"""Script to generate the project's credits."""

from __future__ import annotations

import os
import sys
from collections import defaultdict
from collections.abc import Iterable
from importlib.metadata import distributions
from itertools import chain
from pathlib import Path
from textwrap import dedent

from jinja2 import StrictUndefined
from jinja2.sandbox import SandboxedEnvironment
from packaging.requirements import Requirement

# YORE: EOL 3.10: Replace block with line 2.
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

project_dir = Path(os.getenv("MKDOCS_CONFIG_DIR", "."))
with project_dir.joinpath("pyproject.toml").open("rb") as pyproject_file:
    pyproject = tomllib.load(pyproject_file)
project = pyproject["project"]
project_name = project["name"]
devdeps = [
    dep
    for group in pyproject["dependency-groups"].values()
    for dep in group
    if not dep.startswith("-e")
]

PackageMetadata = dict[str, str | Iterable[str]]
Metadata = dict[str, PackageMetadata]


def _merge_fields(metadata: dict) -> PackageMetadata:
    fields = defaultdict(list)
    for header, value in metadata.items():
        fields[header.lower()].append(value.strip())
    return {
        field: value
        if len(value) > 1 or field in ("classifier", "requires-dist")
        else value[0]
        for field, value in fields.items()
    }


def _norm_name(name: str) -> str:
    return name.replace("_", "-").replace(".", "-").lower()


def _requirements(deps: list[str]) -> dict[str, Requirement]:
    return {_norm_name((req := Requirement(dep)).name): req for dep in deps}


def _extra_marker(req: Requirement) -> str | None:
    if not req.marker:
        return None
    try:
        return next(
            marker[2].value
            for marker in req.marker._markers
            if getattr(marker[0], "value", None) == "extra"
        )
    except StopIteration:
        return None


def _get_metadata() -> Metadata:
    metadata = {}
    for pkg in distributions():
        name = _norm_name(pkg.name)  # type: ignore[attr-defined,unused-ignore]
        metadata[name] = _merge_fields(pkg.metadata)  # type: ignore[arg-type]
        metadata[name]["spec"] = set()
        metadata[name]["extras"] = set()
        metadata[name].setdefault("summary", "")
        _set_license(metadata[name])
    return metadata


def _set_license(metadata: PackageMetadata) -> None:
    license_field = metadata.get("license-expression", metadata.get("license", ""))
    license_name = (
        license_field if isinstance(license_field, str) else " + ".join(license_field)
    )
    check_classifiers = license_name in (
        "UNKNOWN",
        "Dual License",
        "",
    ) or license_name.count("\n")
    if check_classifiers:
        license_names = [
            classifier.rsplit("::", 1)[1].strip()
            for classifier in metadata["classifier"]
            if classifier.startswith("License ::")
        ]
        license_name = " + ".join(license_names)
    metadata["license"] = license_name or "?"


def _get_deps(base_deps: dict[str, Requirement], metadata: Metadata) -> Metadata:
    deps = {}
    for dep_name, dep_req in base_deps.items():
        if dep_name not in metadata or dep_name == "griffe":
            continue
        metadata[dep_name]["spec"] |= {str(spec) for spec in dep_req.specifier}  # type: ignore[operator]
        metadata[dep_name]["extras"] |= dep_req.extras  # type: ignore[operator]
        deps[dep_name] = metadata[dep_name]

    again = True
    while again:
        again = False
        for pkg_name in metadata:
            if pkg_name in deps:
                for pkg_dependency in metadata[pkg_name].get("requires-dist", []):
                    requirement = Requirement(pkg_dependency)
                    dep_name = _norm_name(requirement.name)
                    extra_marker = _extra_marker(requirement)
                    if (
                        dep_name in metadata
                        and dep_name not in deps
                        and dep_name != project["name"]
                        and (
                            not extra_marker or extra_marker in deps[pkg_name]["extras"]
                        )
                    ):
                        metadata[dep_name]["spec"] |= {
                            str(spec) for spec in requirement.specifier
                        }  # type: ignore[operator]
                        deps[dep_name] = metadata[dep_name]
                        again = True

    return deps


def _render_credits() -> str:
    metadata = _get_metadata()
    dev_dependencies = _get_deps(_requirements(devdeps), metadata)
    prod_dependencies = _get_deps(
        _requirements(
            chain(  # type: ignore[arg-type]
                project.get("dependencies", []),
                chain(*project.get("optional-dependencies", {}).values()),
            ),
        ),
        metadata,
    )

    template_data = {
        "project_name": project_name,
        "prod_dependencies": sorted(
            prod_dependencies.values(), key=lambda dep: str(dep["name"]).lower()
        ),
        "dev_dependencies": sorted(
            dev_dependencies.values(), key=lambda dep: str(dep["name"]).lower()
        ),
    }
    template_text = dedent(
        """
        # Credits

        These projects were used to build *{{ project_name }}*. **Thank you!**

        [Python](https://www.python.org/) |
        [uv](https://github.com/astral-sh/uv)

        {% macro dep_line(dep) -%}
        [{{ dep.name }}](https://pypi.org/project/{{ dep.name }}/) | {{ dep.summary }} | {{ ("`" ~ dep.spec|sort(reverse=True)|join(", ") ~ "`") if dep.spec else "" }} | `{{ dep.version }}` | {{ dep.license }}
        {%- endmacro %}

        {% if prod_dependencies -%}
        ### Runtime dependencies

        Project | Summary | Version (accepted) | Version (last resolved) | License
        ------- | ------- | ------------------ | ----------------------- | -------
        {% for dep in prod_dependencies -%}
        {{ dep_line(dep) }}
        {% endfor %}

        {% endif -%}
        {% if dev_dependencies -%}
        ### Development dependencies

        Project | Summary | Version (accepted) | Version (last resolved) | License
        ------- | ------- | ------------------ | ----------------------- | -------
        {% for dep in dev_dependencies -%}
        {{ dep_line(dep) }}
        {% endfor %}

        {% endif -%}
        """,
    )
    jinja_env = SandboxedEnvironment(undefined=StrictUndefined)
    return jinja_env.from_string(template_text).render(**template_data)


print(_render_credits())
